package dev.pfaff.altargui.render;

import dev.pfaff.altargui.shader.BuiltInstanceLayout;
import dev.pfaff.altargui.shader.Program;
import dev.pfaff.altargui.shader.Programs;
import dev.pfaff.altargui.shader.ShaderProg2D;
import dev.pfaff.glass.opengl.enums.GLBlendEquationModeEXT;
import dev.pfaff.glass.opengl.enums.GLBlendingFactor;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;
import dev.pfaff.glass.opengl.enums.GLEnableCap;
import it.unimi.dsi.fastutil.ints.IntArrayList;

import java.lang.foreign.MemorySegment;
import java.lang.foreign.SequenceLayout;
import java.util.List;
import java.util.Spliterator;

import static dev.pfaff.altargui.GuiGlobals.GL;
import static dev.pfaff.altargui.render.GLFlags.ALWAYS_UNBIND;
import static java.lang.foreign.MemoryLayout.sequenceLayout;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public abstract class LayeredRenderBuffer implements AutoCloseable {
//	private static final InstanceLayout CONTEXTUAL_BUFFER_SPEC = new InstanceLayout(List.of(
////			new VertexAttribSpec("ClipRect", VertexAttrib.Type.USHORT, 4)
//			new VertexAttribSpec("ClipRect", VertexAttrib.Type.FLOAT, 4)
//	), VertexAttrib.DIVISOR_PER_INSTANCE);

	private static final SequenceLayout ENTRY_LAYOUT = sequenceLayout(4, JAVA_INT);

	protected final List<Program<ShaderProg2D>> programs = List.of(Programs.RECT_RENDER_LAYER,
			Programs.TEXTURED_RECT_RENDER_LAYER,
			Programs.TEXT_RENDER_LAYER);
	// FIXME: need to rebuild on shader reload
	protected final List<BuiltInstanceLayout> instanceLayouts =
			this.programs.stream().map(layer -> layer.instanceLayout().build(layer.shader().get())).toList();

	/**
	 * Entries are stored as pairs of {@code (program, data1, data2, data3)}
	 */
	private final IntArrayList entries = new IntArrayList();
	protected final RawBuffer instancesBuffer;
	private int currentProgram = -1;
	private boolean uploading = false;

	public LayeredRenderBuffer() {
		this.instancesBuffer = RawBuffer.create(GLBufferTarget.ARRAY_BUFFER, true);
	}

	protected abstract void onRecreate();

	/**
	 * @param capacity the maximum number of bytes that will be uploaded.
	 */
	public void beginUploading(int capacity) {
		if (GLFlags.CHECKS) {
			if (this.uploading) throw new IllegalStateException("Already uploading!");
			this.uploading = true;
		}
		this.entries.clear();
		this.instancesBuffer.reset(capacity, this::onRecreate);
		this.instancesBuffer.beginUploading();
	}

	public final void finishUploading() {
		if (GLFlags.CHECKS) {
			if (!this.uploading) throw new IllegalStateException("Not uploading!");
			this.uploading = false;
		}
		this.endCurrentLayer();
		this.instancesBuffer.finishUploading();
	}

	@Override
	public void close() {
		this.instancesBuffer.close();
	}

	/**
	 * @param ms a buffer of 3 integers. Store whatever you need here.
	 * @param program the shader program used for the layer.
	 * @param start the byte offset of the start of the program's vertex data.
	 * @param end the byte offset of the end of the program's vertex data.
	 */
	protected abstract void putDataForEntry(MemorySegment ms, int program, int start, int end);

	/**
	 * @param ms a buffer of 3 integers.
	 * @param layer
	 */
	protected abstract VAO getVAOForEntry(MemorySegment ms, int layer);

	/**
	 * The VAO will be bound by the caller.
	 *
	 * @param ms a buffer of 3 integers.
	 * @param layer
	 */
	protected abstract void drawEntry(MemorySegment ms, int layer, Program<?> layerInfo);

	private void useProgram(int program) {
		if (this.currentProgram != program) {
			this.endCurrentLayer();
			this.currentProgram = program;
			this.entries.add(program);
			var stride = this.instanceLayouts.get(program).stride();
			RawBuffer instancesBuffer = this.instancesBuffer;
			var len = instancesBuffer.length();
			var index = (len + (stride - 1)) / stride;
			var alignedLen = index * stride;
			instancesBuffer.setLength(instancesBuffer.length() + (alignedLen - len));
			this.entries.add(index);
//			GuiGlobals.LOGGER.log(Level.INFO, () -> STR."began program \{program} at \{len} aligned up to \{alignedLen} (index \{index})");
		}
	}

	public final boolean endCurrentLayer() {
		int program = this.currentProgram;
		if (program == -1) return false;
		int index = this.instancesBuffer.length() / this.instanceLayouts.get(program).stride();
		int startIndex = this.entries.getLast();
		this.entries.size(this.entries.size() + 2);
		this.putDataForEntry(
				MemorySegment.ofArray(this.entries.elements())
						.asSlice((long)Integer.BYTES * (this.entries.size() - 3),Integer.BYTES * 3),
				program,
				startIndex,
				index);
//			GuiGlobals.LOGGER.log(Level.INFO, () -> STR."ended program \{program} at \{this.instancesBuffer.length()} (index \{index})");
		this.currentProgram = -1;
		return true;
	}

	public final void iSpecUpload(RenderObject object, int program) {
//		Logger.log(() -> STR."uploading \{object}");
		this.useProgram(program);
		this.instancesBuffer.upload(object);
		if (GLFlags.CHECKS) {
			var layout = this.instanceLayouts.get(program);
			var stride = layout.stride();
			var trailing = this.instancesBuffer.length() % stride;
			if (trailing != 0) {
				throw new BufferOverflowException(STR."""
				Upload of \{object} left \{trailing} trailing bytes in buffer.
				  This is either an underflow or an overflow.
				  The object consists of \{object.instanceCount()} instances (\{object.size()} bytes)
				  The object layout is \{object.instanceLayout()} (\{object.instanceLayout().byteSize()} bytes)
				  The layer has a stride of \{stride} bytes
				  The layer id is \{program}
				  The layer layout is \{layout}""");
			}
		}
	}

	public final void upload(RenderObject object) {
		object.uSpecUploadLayered(this);
	}

	public final class Drawer implements AutoCloseable {
		private final Spliterator<MemorySegment> entries;

		private Drawer(Spliterator<MemorySegment> entries) {
			this.entries = entries;
		}

		/**
		 * Draws the next program.
		 */
		public boolean drawNext() {
			return this.entries.tryAdvance(entry -> {
				int program = entry.getAtIndex(JAVA_INT, 0);
//				int start = iter.nextInt();
//				int end = iter.nextInt();
//				int instanceCount = end - start;

				var layerInfo = LayeredRenderBuffer.this.programs.get(program);

				var shader = layerInfo.shader().get();

				var entryData = entry.asSlice(Integer.BYTES);

				var vao = LayeredRenderBuffer.this.getVAOForEntry(entryData, program);
				vao.bind();

				layerInfo.drawSetup().accept(shader);
				GL.checkError();

				shader.bind();
				GL.checkError();

				LayeredRenderBuffer.this.drawEntry(entryData, program, layerInfo);

//				GuiGlobals.LOGGER.log(Level.INFO, () -> STR."drew program \{program} with \{shader.getName()}");
			});
		}

		public void drawRemaining() {
			while (drawNext()) {}
		}

		@Override
		public void close() {
			if (ALWAYS_UNBIND) GLUtil.glxUnbindVertexArray();
		}
	}

	public final Drawer drawer() {
		GL.glEnable(GLEnableCap.BLEND);
		GL.glBlendEquation(GLBlendEquationModeEXT.FUNC_ADD);
		GL.glBlendFunc(GLBlendingFactor.SRC_ALPHA, GLBlendingFactor.ONE_MINUS_SRC_ALPHA);
		GL.glDisable(GLEnableCap.DEPTH_TEST);

		var buf = MemorySegment.ofArray(this.entries.elements()).asSlice(0, (long) Integer.BYTES * this.entries.size());
		if (GLFlags.CHECKS && this.entries.size() % 4 != 0) throw new IllegalStateException();
		return new Drawer(buf.spliterator(ENTRY_LAYOUT));
	}

	public final int currentProgram() {
		return this.currentProgram;
	}

	public final int byteCount() {
		return this.instancesBuffer.length();
	}
}
